@@ -249,11 +249,6 @@ class Dumper(DumperBase):
             if bitsize == 0:
                 bitsize = f.GetType().GetByteSize() * 8
             bitpos = f.GetOffsetInBits()
-            # Correction for some bitfields. Size 0 can occur for
-            # types without debug information.
-            if bitsize > 0:
-                #bitpos = bitpos % bitsize
-                bitpos = bitpos % 8 # Reported type is always wrapping type!
             fieldBits[f.name] = (bitsize, bitpos, f.IsBitfield())
 
         # Normal members and non-empty base classes.
@@ -269,7 +264,7 @@ class Dumper(DumperBase):
                 (fieldBitsize, fieldBitpos, isBitfield) = fieldBits[fieldName]
             else:
                 fieldBitsize = nativeFieldType.GetByteSize() * 8
-                fieldBitpost = None
+                fieldBitpos = None
                 isBitfield = False
 
             if isBitfield: # Bit fields
@@ -445,8 +440,8 @@ class Dumper(DumperBase):
                     warn('UNKNOWN TYPE KEY: %s: %s' % (typeName, code))
             elif code == lldb.eTypeClassEnumeration:
                 tdata.code = TypeCodeEnum
-                tdata.enumDisplay = lambda intval, addr : \
-                    self.nativeTypeEnumDisplay(nativeType, intval)
+                tdata.enumDisplay = lambda intval, addr, form : \
+                    self.nativeTypeEnumDisplay(nativeType, intval, form)
             elif code in (lldb.eTypeClassComplexInteger, lldb.eTypeClassComplexFloat):
                 tdata.code = TypeCodeComplex
             elif code in (lldb.eTypeClassClass, lldb.eTypeClassStruct, lldb.eTypeClassUnion):
@@ -457,7 +452,7 @@ class Dumper(DumperBase):
                     self.listMembers(value, nativeType)
                 tdata.templateArguments = self.listTemplateParametersHelper(nativeType)
             elif code == lldb.eTypeClassFunction:
-                tdata.code = TypeCodeFunction,
+                tdata.code = TypeCodeFunction
             elif code == lldb.eTypeClassMemberPointer:
                 tdata.code = TypeCodeMemberPointer
 
@@ -534,17 +529,32 @@ class Dumper(DumperBase):
         #warn('NATIVE TYPE ID FOR %s IS %s' % (name, typeId))
         return typeId
 
-    def nativeTypeEnumDisplay(self, nativeType, intval):
+    def nativeTypeEnumDisplay(self, nativeType, intval, form):
         if hasattr(nativeType, 'get_enum_members_array'):
+            enumerators = []
+            flags = []
+            found = False
             for enumMember in nativeType.get_enum_members_array():
                 # Even when asking for signed we get unsigned with LLDB 3.8.
-                diff = enumMember.GetValueAsSigned() - intval
-                mask = (1 << nativeType.GetByteSize() * 8) - 1
-                if diff & mask == 0:
-                    path = nativeType.GetName().split('::')
-                    path[-1] = enumMember.GetName()
-                    return '%s (%d)' % ('::'.join(path), intval)
-        return '%d' % intval
+                value = enumMember.GetValueAsSigned()
+                name = nativeType.GetName().split('::')
+                name[-1] = enumMember.GetName()
+                if value == intval:
+                    return '::'.join(name) + ' (' + (form % intval) + ')'
+                enumerators.append(('::'.join(name), value))
+
+            given = intval
+            for (name, value) in enumerators:
+                if value & given != 0:
+                    flags.append(name)
+                    given = given & ~value
+                    found = True
+
+            if not found or given != 0:
+                flags.append('unknown: %d' % given)
+
+            return '(' + ' | '.join(flags) + ') (' + (form % intval) + ')'
+        return form % intval
 
     def nativeDynamicTypeName(self, address, baseType):
         return None # FIXME: Seems sufficient, no idea why.
@@ -649,6 +659,21 @@ class Dumper(DumperBase):
     def isMsvcTarget(self):
         return False
 
+    def prettySymbolByAddress(self, address):
+        try:
+            result = lldb.SBCommandReturnObject()
+            # Cast the address to a function pointer to get the name and location of the function.
+            expression = 'po (void (*)()){}'
+            self.debugger.GetCommandInterpreter().HandleCommand(expression.format(address), result)
+            output = ''
+            if result.Succeeded():
+                output = result.GetOutput().strip()
+            if output:
+                return output
+        except:
+            pass
+        return '0x%x' % address
+
     def qtVersionAndNamespace(self):
         for func in self.target.FindFunctions('qVersion'):
             name = func.GetSymbol().GetName()
@@ -774,6 +799,9 @@ class Dumper(DumperBase):
             if typeobj is not None:
                 return typeobj
 
+        return self.lookupNativeTypeInAllModules(name)
+
+    def lookupNativeTypeInAllModules(self, name):
         needle = self.canonicalTypeName(name)
         #warn('NEEDLE: %s ' % needle)
         warn('Searching for type %s across all target modules, this could be very slow' % name)
@@ -812,8 +840,7 @@ class Dumper(DumperBase):
         self.startMode_ = args.get('startmode', 1)
         self.breakOnMain_ = args.get('breakonmain', 0)
         self.useTerminal_ = args.get('useterminal', 0)
-        self.processArgs_ = args.get('processargs', [])
-        self.processArgs_ = list(map(lambda x: self.hexdecode(x), self.processArgs_))
+        self.processArgs_ = self.hexdecode(args.get('processargs'))
         self.environment_ = args.get('environment', [])
         self.environment_ = list(map(lambda x: self.hexdecode(x), self.environment_))
         self.attachPid_ = args.get('attachpid', 0)
@@ -905,9 +932,24 @@ class Dumper(DumperBase):
         elif self.startMode_ == AttachCore:
             coreFile = args.get('coreFile', '');
             self.process = self.target.LoadCore(coreFile)
-            self.reportState('enginerunokandinferiorunrunnable')
+            if self.process.IsValid():
+                self.reportState('enginerunokandinferiorunrunnable')
+            else:
+                self.reportState('enginerunfailed')
         else:
-            launchInfo = lldb.SBLaunchInfo(self.processArgs_)
+            # This does not seem to work on Linux nor macOS?
+            #launchInfo = lldb.SBLaunchInfo([self.processArgs_])
+            #launchInfo.SetShellExpandArguments(True)
+            args = []
+            try:
+                import subprocess
+                cmd = 'for x in {} ; do printf "%s\n" "$x" ; done' \
+                    .format(self.processArgs_)
+                args = subprocess.check_output(cmd, shell=True, cwd=self.workingDirectory_).split()
+            except:
+                # Wrong, but...
+                args = self.processArgs_
+            launchInfo = lldb.SBLaunchInfo(args)
             launchInfo.SetWorkingDirectory(self.workingDirectory_)
             launchInfo.SetEnvironmentEntries(self.environment_, False)
             if self.breakOnMain_:
@@ -1041,6 +1083,7 @@ class Dumper(DumperBase):
             addr = frame.GetPCAddress().GetLoadAddress(self.target)
 
             functionName = frame.GetFunctionName()
+            module = frame.GetModule()
 
             if isNativeMixed and functionName == '::qt_qmlDebugMessageAvailable()':
                 interpreterStack = self.extractInterpreterStack()
@@ -1060,6 +1103,7 @@ class Dumper(DumperBase):
             result += ',address="0x%x"' % addr
             result += ',function="%s"' % functionName
             result += ',line="%d"' % lineNumber
+            result += ',module="%s"' % module
             result += ',file="%s"},' % fileName
         result += ']'
         result += ',hasmore="%d"' % isLimited
@@ -1106,6 +1150,13 @@ class Dumper(DumperBase):
 
         self.setVariableFetchingOptions(args)
 
+        # Reset certain caches whenever a step over / into / continue
+        # happens.
+        # FIXME: Caches are currently also cleared if currently
+        # selected frame is changed, that shouldn't happen.
+        if not self.partialVariable:
+            self.resetPerStepCaches()
+
         anyModule = self.target.GetModuleAtIndex(0)
         anySymbol = anyModule.GetSymbolAtIndex(0)
         self.fakeAddress = int(anySymbol.GetStartAddress())
@@ -1148,7 +1199,7 @@ class Dumper(DumperBase):
         #    values = [frame.FindVariable(partialVariable)]
         #else:
         if True:
-            values = list(frame.GetVariables(True, True, False, False))
+            values = list(frame.GetVariables(True, True, False, True))
             values.reverse() # To get shadowed vars numbered backwards.
 
         variables = []
@@ -1446,6 +1497,21 @@ class Dumper(DumperBase):
                 bp.SetOneShot(bool(args['oneshot']))
         self.reportResult(self.describeBreakpoint(bp), args)
 
+    def enableSubbreakpoint(self, args):
+        lldbId = int(args['lldbid'])
+        locId = int(args['locid'])
+        bp = self.target.FindBreakpointByID(lldbId)
+        res = False
+        enabled = False
+        if bp.IsValid():
+            loc = bp.FindLocationByID(locId)
+            if loc.IsValid():
+                loc.SetEnabled(bool(args['enabled']))
+                enabled = loc.IsEnabled()
+                res = True
+        self.reportResult('success="%s",enabled="%s",locid="%s"'
+                        % (int(res), int(enabled), locId), args)
+
     def removeBreakpoint(self, args):
         lldbId = int(args['lldbid'])
         if lldbId > qqWatchpointOffset:
@@ -1499,7 +1565,7 @@ class Dumper(DumperBase):
         self.reportResult('', args)
 
     def executeNextI(self, args):
-        self.currentThread().StepInstruction(lldb.eOnlyThisThread)
+        self.currentThread().StepInstruction(True)
         self.reportResult('', args)
 
     def executeStep(self, args):
@@ -1508,22 +1574,20 @@ class Dumper(DumperBase):
 
     def shutdownInferior(self, args):
         self.isShuttingDown_ = True
-        if self.process is None:
-            self.reportState('inferiorshutdownok')
-        else:
+        if self.process is not None:
             state = self.process.GetState()
             if state == lldb.eStateStopped:
                 self.process.Kill()
-            self.reportState('inferiorshutdownok')
+        self.reportState('inferiorshutdownfinished')
         self.reportResult('', args)
 
     def quit(self, args):
-        self.reportState('engineshutdownok')
+        self.reportState('engineshutdownfinished')
         self.process.Kill()
         self.reportResult('', args)
 
     def executeStepI(self, args):
-        self.currentThread().StepInstruction(lldb.eOnlyThisThread)
+        self.currentThread().StepInstruction(False)
         self.reportResult('', args)
 
     def executeStepOut(self, args = {}):
@@ -1585,13 +1649,12 @@ class Dumper(DumperBase):
 
     def activateFrame(self, args):
         self.reportToken(args)
-        thread = args['thread']
         self.currentThread().SetSelectedFrame(args['index'])
         self.reportResult('', args)
 
     def selectThread(self, args):
         self.reportToken(args)
-        self.process.SetSelectedThreadByID(args['id'])
+        self.process.SetSelectedThreadByID(int(args['id']))
         self.reportResult('', args)
 
     def fetchFullBacktrace(self, _ = None):
@@ -1694,13 +1757,39 @@ class Dumper(DumperBase):
         value = frame.FindVariable(exp)
         return value
 
+    def setValue(self, address, typename, value):
+        sbtype = self.lookupNativeType(typename)
+        error = lldb.SBError()
+        sbaddr = lldb.SBAddress(address, self.target)
+        sbvalue = self.target.CreateValueFromAddress('x', sbaddr, sbtype)
+        sbvalue.SetValueFromCString(str(value), error)
+
+    def setValues(self, address, typename, values):
+        sbtype = self.lookupNativeType(typename)
+        sizeof = sbtype.GetByteSize()
+        error = lldb.SBError()
+        for i in range(len(values)):
+            sbaddr = lldb.SBAddress(address + i * sizeof, self.target)
+            sbvalue = self.target.CreateValueFromAddress('x', sbaddr, sbtype)
+            sbvalue.SetValueFromCString(str(values[i]), error)
+
     def assignValue(self, args):
         self.reportToken(args)
         error = lldb.SBError()
-        exp = self.hexdecode(args['exp'])
+        expr = self.hexdecode(args['expr'])
         value = self.hexdecode(args['value'])
-        lhs = self.findValueByExpression(exp)
-        lhs.SetValueFromCString(value, error)
+        simpleType = int(args['simpleType'])
+        lhs = self.findValueByExpression(expr)
+        typeName = lhs.GetType().GetName()
+        typeName = typeName.replace('::', '__')
+        pos = typeName.find('<')
+        if pos != -1:
+            typeName = typeName[0:pos]
+        if typeName in self.qqEditable and not simpleType:
+            expr = self.parseAndEvaluate(expr)
+            self.qqEditable[typeName](self, expr, value)
+        else:
+            lhs.SetValueFromCString(value, error)
         self.reportResult(self.describeError(error), args)
 
     def watchPoint(self, args):
@@ -1878,6 +1967,10 @@ class SummaryDumper(Dumper, LogMixin):
     def report(self, stuff):
         return # Don't mess up lldb output
 
+    def lookupNativeTypeInAllModules(self, name):
+        warn('Failed to resolve type %s' % name)
+        return None
+
     def dump_summary(self, valobj, expanded = False):
         try:
             from pygdbmi import gdbmiparser
